val languagesAndAuthors =
  mapOf("Java" to "Gosling", "Lisp" to "McCarthy", "Ruby" to "Matz")

val xml = elem("languages") {
  languagesAndAuthors.forEach { (title, author) ->
    elem("language", "title" to title) {
      elem("author") {
        text(author)
      }
    }
  }
}

println(xml)

fun elem(name: String, vararg attributes: Pair<String, String>,
  block: XmlNode.() -> Unit = {}): XmlNode = 
    XmlNode(name, *attributes).apply { block() }

class XmlNode(
  val elementName: String, vararg attributes: Pair<String, String>) {

  var textValue = ""
  val children = mutableListOf<XmlNode>()
  val attributes = attributes
  
  fun elem(name: String, vararg attributes: Pair<String, String>,
    block: XmlNode.() -> Unit = {}) =
      children.add(XmlNode(name, *attributes).apply { block() })
  
  fun text(value: String) {
    textValue = value
  }

  override fun toString() = xmlAsString()
  private fun xmlAsString(indentation: String = ""): String =
    listOf(
        renderOpeningTag(indentation),
        renderTextNode(indentation),
        renderChildren(indentation),
        renderClosingTag(indentation))
      .filter(String::isNotEmpty)
      .joinToString("\n")
  
  private fun renderOpeningTag(indentation: String): String {
    val attributeValues = attributes.map { (title, author) -> "$title='$author'" }
      .joinToString(" ", prefix = " ")
    
    return "$indentation<${(elementName + attributeValues).trim()}>"
  }
  
  private fun renderTextNode(indentation: String) =
    if(textValue.isEmpty()) "" else " $indentation$textValue"
  
  private fun renderChildren(indentation: String) = 
    children.map { node -> "${node.xmlAsString("$indentation ")}"}
      .joinToString("\n")
      
  private fun renderClosingTag(indentation: String) = "$indentation</$elementName>"
}
